import json
import os
import threading

import mosaik_api_v3
import numpy as np
import pandas as pd
from midas.util.logging import set_and_init_logger

from . import LOG
from .meta import META

pd.set_option("io.hdf.default_format", "table")


class MidasHdf5(mosaik_api_v3.Simulator):
    """A mosaik data collector using pytables and hdf.

    This mosaik simulator is able to take any number of inputs from
    mosaik and stores them in a HDF database file. This is done using
    pandas and pytables.

    The only requirement is that the input can be put into a cell of a
    pandas dataframe, i.e., basic data types like int, float, and
    string are supported. More complex structures need to be serialized
    beforehand, e.g., if you want to store a dictionary, use json to
    serialize it to a string.

    MidasHdf5 tries to be non-destructive when it comes to the database
    file. If you pass a filename that already exists on the table, a
    new filename will be searched. This is usually just the old
    filename with an increment, e.g., if you pass my_db.hdf5, which
    already exists, the filename will be my_db-2.hdf5. This behavior
    can be disabled by providing the flag *overwrite*. However, even
    then the old database file will not directly overwritten, but moved
    to my_db.hdf5.old instead.

    Another option is *buffer_size*. Running large simulations can lead
    to a really large data set that is stored in-memory. This may be
    undesired for various reasons. Setting *buffer_size* to a value >0,
    the collected data will be saved to disc every buffer_size steps.
    Generally, writing to disc is done with a thread so using the
    buffer size should not affect the overall performance. It is
    recommended, though, to not set the buffer_size too low.

    """

    def __init__(self):
        super().__init__(META)

        self.sid = None
        self.eid = "Database-0"
        self.database = None
        self.filename = None
        self.step_size = None
        self.current_size = 0
        self.buffer_size = None
        self.saved_rows = 0
        self.finalized = False
        self.keep_old_files = True
        self._worker = None

    def init(self, sid, **sim_params):
        self.sid = sid
        self.step_size = sim_params.get("step_size", 900)

        return self.meta

    def create(self, num, model, **model_params):
        errmsg = (
            "You should realy not try to instantiate more than one "
            "database. If your need another database, create a new "
            "simulator as well."
        )
        assert self.database is None, errmsg
        assert num == 1, errmsg

        self.keep_old_files = model_params.get("keep_old_files", False)
        self.filename = model_params.get("filename", None)
        if self.filename is not None and not self.filename.endswith(".hdf5"):
            self.filename = f"{self.filename}.hdf5"

        if self.keep_old_files:
            LOG.debug(
                "Keep_old_files is set to True. Attempting to find a unique "
                "filename for the database."
            )
            incr = 2
            new_filename = self.filename
            while os.path.exists(new_filename):
                new_filename = f"{self.filename[:-5]}_{incr}.hdf5"
                incr += 1
            self.filename = new_filename
        elif os.path.exists(self.filename):
            os.rename(self.filename, f"{self.filename}.old")

        self.buffer_size = model_params.get("buffer_size", 0)

        if self.buffer_size is None:
            self.buffer_size = 0
            LOG.info(
                "Buffer size is set to 0. Store will be saved as-whole"
                "once the simulation is finished."
            )

        LOG.info("Saving results to database at '%s'.", self.filename)
        self.database = dict()

        return [{"eid": self.eid, "type": model}]

    def step(self, time, inputs, max_advance=0):
        data = inputs.get(self.eid, {})

        if not data:
            LOG.info(
                "Did not receive any inputs. "
                "Did you connect anything to the store?"
            )

        current = dict()
        for attr, src_ids in data.items():
            for src_id, val in src_ids.items():
                sid, eid = src_id.split(".")
                key = f"{eid}___{attr}".replace("-", "__")
                sid = sid.replace("-", "__")

                current.setdefault(sid, dict())
                current[sid].setdefault("cols", list()).append(key)
                if isinstance(val, (list, dict, np.ndarray)):
                    val = json.dumps(val)
                elif isinstance(val, pd.DataFrame):
                    val = val.to_json()
                current[sid].setdefault("vals", list()).append(val)

        if self._worker is not None:
            LOG.debug("Waiting for the store worker to finish...")
            self._worker.join()
            LOG.debug("Clearing current database.")
            self.database = dict()
            self._worker = None

        for sid, data in current.items():
            self.database.setdefault(sid, pd.DataFrame())

            ndf = pd.DataFrame([data["vals"]], columns=data["cols"])

            self.database[sid] = pd.concat(
                [self.database[sid], ndf], ignore_index=True
            )

        if self.buffer_size > 0:
            self.current_size += 1

            if self.current_size >= self.buffer_size:
                self._clear_buffer()
                self.current_size = 0

        return time + self.step_size

    def get_data(self, outputs):
        return dict()

    def finalize(self):
        if self.finalized:
            return
        else:
            self.finalized = True

        append = self.buffer_size > 0
        self._to_hdf(append)

    def _clear_buffer(self):
        LOG.debug("Starting worker thread to save the current database...")
        self._worker = threading.Thread(target=self._to_hdf)
        self._worker.start()

    def _to_hdf(self, append=True):
        if not self.database:
            LOG.warning("Database is empty. Unable to write anything to disk.")
            return

        errors = list()
        for sid, data in self.database.items():
            try:
                data.index += self.saved_rows
                data.to_hdf(
                    self.filename, key=sid, format="table", append=append
                )
            except Exception:
                LOG.info(
                    "Couldn't save data of simulator %s. Trying to load "
                    "existing data and append manually. ",
                    sid,
                )
                try:
                    edata = pd.read_hdf(self.filename, key=sid)
                    edata = pd.concat([edata, data])
                    edata.to_hdf(
                        self.filename, key=sid, format="table", append=False
                    )
                    LOG.info(
                        "Successfully appended the data. One reason could be "
                        "that some values have inconsistent types, e.g., are "
                        "int in the one step and float in the other. As long "
                        "as this is not fixed, this message will probably "
                        "re-appear."
                    )
                except Exception as err:
                    LOG.error(
                        "Could not append data for simulator %s. "
                        "Attempting to export the data to csv.",
                        sid,
                    )
                    fname = os.path.abspath(
                        os.path.join(os.getcwd(), f"fail_{sid}.csv")
                    )
                    LOG.debug("Failed at SID '%s'.", sid)
                    LOG.debug("Columns: %s ", data.columns)
                    LOG.debug("Trying to export data to '%s'", fname)
                    LOG.debug("%s", data.info())

                    data.to_csv(fname)

                    errors.append(err)

        current_num_rows = data.shape[0]
        self.saved_rows += current_num_rows
        if len(errors) > 0:
            LOG.warning("Worker finished with errors: %s", errors)
        else:
            LOG.debug("Worker finished.")
        LOG.info("Wrote %d rows into the store.", current_num_rows)


if __name__ == "__main__":
    set_and_init_logger(0, "store-logfile", "midas-store.log", replace=True)

    mosaik_api_v3.start_simulation(MidasHdf5())



META = {
    "type": "time-based",
    "models": {
        "Database": {
            "public": True,
            "any_inputs": True,
            "params": ["filename", "verbose", "buffer_size", "keep_old_files"],
            "attrs": [],
        }
    },
}



import os
import unittest

import numpy as np
import pandas as pd

from midas_store.simulator import MidasHdf5


class TestMidasHdf5(unittest.TestCase):
    def setUp(self):
        self.inputs1 = {
            "Database-0": {
                "p_mw": {
                    "DummySim-0.DummyHousehold-0": 0.1,
                    "DummySim-0.DummyHousehold-1": 0.2,
                    "DummySim-1.DummyPV-0": 0.3,
                },
                "q_mvar": {
                    "DummySim-0.DummyHousehold-0": 0.01,
                    "DummySim-0.DummyHousehold-1": 0.02,
                    "DummySim-1.DummyPV-0": 0.03,
                },
                "t_air": {"DummyWeather-0.WeatherCurrent-0": 15.0},
                "schedule": {"DummySim-2.DummyCHP-0": [10, 12, 10]},
            }
        }
        self.inputs2 = {
            "Database-0": {
                "p_mw": {
                    "DummySim-0.DummyHousehold-0": 0.02,
                    "DummySim-0.DummyHousehold-1": 0.02,
                    "DummySim-1.DummyPV-0": 0.03,
                    "DummySim-2.DummyCHP-0": 0.5,
                },
                "q_mvar": {
                    "DummySim-0.DummyHousehold-0": 0.01,
                    "DummySim-0.DummyHousehold-1": 0.015,
                    "DummySim-1.DummyPV-0": 0.01,
                },
                "t_air": {"DummyWeather-0.WeatherCurrent-0": 15.0},
                "wind": {"DummyWeather-1.WeatherForecast-0": 20},
            }
        }

    def test_setup(self):
        """Test store creation and ensure to allow only one instance."""
        dbsim = MidasHdf5()

        dbsim.init("MidasHdf5", step_size=900)

        # Only one instance allowed
        with self.assertRaises(AssertionError):
            dbsim.create(2, "Database", filename="there.hdf5")

        dbsim.create(1, "Database", filename="here.hdf5")

        self.assertIsNotNone(dbsim.database)
        self.assertEqual("here.hdf5", dbsim.filename)
        self.assertEqual(900, dbsim.step_size)

        # Only one instance allowed
        with self.assertRaises(AssertionError):
            dbsim.create(1, "Database", filename="there.hdf5")

    def test_step(self):
        dbsim = MidasHdf5()
        dbsim.init("MidasHdf5", step_size=900)
        dbsim.create(1, "Database", filename="here.hdf5")
        dbsim.step(0, self.inputs1)

        dbsim.step(900, self.inputs2)

        # print(dbsim.database["DummySim-0"])

    def test_step_no_inputs(self):
        dbfile = "not-there.hdf5"
        dbsim = MidasHdf5()
        dbsim.init("MidasHdf5", step_size=900)
        dbsim.create(1, "Database", filename=dbfile)
        dbsim.step(0, {})
        with self.assertLogs("midas_store", level="WARNING") as cm:
            dbsim.finalize()

        self.assertIn("Database is empty.", cm.output[0])
        self.assertFalse(os.path.exists(dbfile))

    @unittest.skip
    def test_huge_dataset(self):
        """Test if a large dataset can be stored. Takes very long
        and should not be necessary most of the time.
        """
        dbsim = MidasHdf5()
        dbsim.init("MidasHdf5", step_size=900)
        dbsim.create(1, "Database", filename="here.hdf5")

        for idx in range(5 * 365 * 24 * 4):
            dbsim.step(idx * 900, self.inputs1)

            if idx % 96 == 0:
                print(idx / 96, end="\r")

        print()
        dbsim.finalize()

    @unittest.skip
    def test_huge_dataset2(self):
        dbsim = MidasHdf5()
        dbsim.init("MidasHdf5", step_size=900)
        dbsim.create(1, "Database", filename="here.hdf5")
        data = np.ones((5 * 365 * 24 * 4, 4))
        dbsim.database["DummySim-0"] = pd.DataFrame(
            data,
            columns=[
                "DummyMod-0.p_mw",
                "DummyMod-0.q_mvar",
                "DummyMod-1.p_mw",
                "DummyMod-1.q_mvar",
            ],
        )
        dbsim.database["DummySim-1"] = pd.DataFrame(
            data,
            columns=[
                "DummyMod-0.p_mw",
                "DummyMod-0.q_mvar",
                "DummyMod-1.p_mw",
                "DummyMod-1.q_mvar",
            ],
        )
        dbsim.database["DummySim-2"] = pd.DataFrame(
            data,
            columns=[
                "DummyMod-0.p_mw",
                "DummyMod-0.q_mvar",
                "DummyMod-1.p_mw",
                "DummyMod-1.q_mvar",
            ],
        )
        dbsim.finalize()
        print("saved", dbsim.database["DummySim-0"].shape)

    def test_threaded_store(self):
        """Test if the use of thread preserves all information."""
        dbfile = os.path.abspath("threads_here.hdf5")
        if os.path.exists(dbfile):
            os.remove(dbfile)

        dbsim = MidasHdf5()
        dbsim.init("MidasHdf5", step_size=900)
        dbsim.create(1, "Database", filename=dbfile, buffer_size=5)

        for step in range(16):
            dbsim.step(step * 900, self.inputs1)

        try:
            dbsim.finalize()
            db_restored = dict()
            for sid in ["DummySim-0", "DummySim-1", "DummyWeather-0"]:
                nsid = sid.replace("-", "__")
                db_restored[nsid] = pd.read_hdf(dbfile, nsid)
                self.assertEqual(16, len(db_restored[nsid].index))

            # Clean up
            os.remove(dbfile)
        except FileNotFoundError:
            print("Db does not exist")

    def test_unthreaded_store(self):
        """Test if the store is still valid even without threads."""
        dbfile = os.path.abspath("no_threads_here.hdf5")
        if os.path.exists(dbfile):
            os.remove(dbfile)

        dbsim = MidasHdf5()
        dbsim.init("MidasHdf5", step_size=900)
        dbsim.create(1, "Database", filename=dbfile, buffer_size=0)

        for step in range(16):
            dbsim.step(step * 900, self.inputs1)

        try:
            dbsim.finalize()

            db_restored = dict()
            for sid in ["DummySim-0", "DummySim-1", "DummyWeather-0"]:
                nsid = sid.replace("-", "__")
                db_restored[nsid] = pd.read_hdf(dbfile, nsid)
                self.assertEqual(16, len(db_restored[nsid].index))

            # Clean up
            os.remove(dbfile)
        except FileNotFoundError:
            print("Db does not exist")


if __name__ == "__main__":
    unittest.main()



"""MIDAS scenario upgrade module.

This module adds a mosaikhdf database to the scenario.

"""
import logging
import os

from midas.scenario.upgrade_module import UpgradeModule

LOG = logging.getLogger(__name__)


class DatabaseModule(UpgradeModule):
    def __init__(self):
        super().__init__(
            module_name="store",
            default_scope_name="database",
            default_sim_config_name="MidasStore",
            default_import_str="midas_store.simulator:MidasHdf5",
            default_cmd_str=(
                "%(python)s -m midas_store.simulator %(addr)s"
            ),
            log=LOG,
        )
        self.default_filename = "midas_store.hdf5"
        self.filename = None

    def check_module_params(self, module_params):
        """Check module params for this upgrade."""

        module_params.setdefault(self.default_scope_name, dict())
        module_params.setdefault("buffer_size", 0)
        module_params.setdefault("keep_old_files", False)
        module_params.setdefault("filename", self.default_filename)
        if module_params["filename"] is not None:
            module_params["filename"] = os.path.abspath(
                os.path.join(
                    self.scenario.base.output_path, module_params["filename"]
                )
            )

    def check_sim_params(self, module_params):

        self.filename = module_params["filename"]

        if "MosaikHdf5" in self.sim_params["import_str"]:
            self.scenario["db_restricted"] = True
            self.sim_params["duration"] = self.scenario["end"]
        else:
            self.sim_params.setdefault(
                "buffer_size", module_params["buffer_size"]
            )
            self.sim_params.setdefault(
                "keep_old_files", module_params["keep_old_files"]
            )

    def start_models(self):
        mod_key = "database"
        params = {"filename": self.filename}

        if "midas" in self.sim_params["import_str"]:
            params["buffer_size"] = self.sim_params["buffer_size"]
            params["keep_old_files"] = self.sim_params["keep_old_files"]

        self.start_model(mod_key, "Database", params)

    def connect(self):
        pass

    def connect_to_db(self):
        pass
